@@ -53,12 +53,13 @@ from abc import ABCMeta, abstractproperty, abstractmethod
 import collections
 import itertools
 import threading
+import Queue
 
 import numpy as np
 import numpy.ma as ma
 
 
-__version__ = '0.1'
+__version__ = '0.2'
 
 
 class Array(object):
@@ -69,6 +70,10 @@ class Array(object):
     """
     __metaclass__ = ABCMeta
 
+    def __repr__(self):
+        return '<{} shape={} dtype={!r}>'.format(type(self).__name__,
+                                                 self.shape, self.dtype)
+
     @property
     def ndim(self):
         """The number of dimensions in this virtual array."""
@@ -110,13 +115,13 @@ class ArrayAdapter(Array):
 
     """
     def __init__(self, concrete, keys=()):
-        # concrete has:
+        # concrete must have:
         #   dtype
-        #   ndim
+        #   shape
         self._concrete = concrete
         if not isinstance(keys, tuple):
             keys = (keys,)
-        assert len(keys) <= concrete.ndim
+        assert len(keys) <= len(concrete.shape)
         result_keys = []
         for axis, (key, size) in enumerate(zip(keys, concrete.shape)):
             result_key = self._cleanup_new_key(key, size, axis)
@@ -244,10 +249,6 @@ class ArrayAdapter(Array):
 
         return ArrayAdapter(self._concrete, tuple(result_keys))
 
-    def __repr__(self):
-        return '<ArrayAdapter shape={} dtype={!r}>'.format(
-            self.shape, self.dtype)
-
     def _apply_keys(self):
         # If we have more than one tuple as a key, then NumPy does
         # "fancy" indexing, instead of "column-based" indexing, so we
@@ -362,10 +363,6 @@ class ArrayStack(Array):
             result = self._stack[stack_keys][item_keys]
         return result
 
-    def __repr__(self):
-        return '<ArrayStack stack_shape={} item_shape={} dtype={!r}>'.format(
-            self._stack.shape, self._item_shape, self.dtype)
-
     def __setitem__(self, keys, value):
         assert len(keys) == self._stack.ndim
         for key in keys:
@@ -515,44 +512,130 @@ class LinearMosaic(Array):
         return data
 
 
-def _process_chunks(a, chunk_handler):
-    #   chunk_size = 2      => 54s ~ 115% CPU
-    #   chunk_size = 10     => 42s ~ 105% CPU (quicker than CDO!)
-    #   chunk_size = 100    => 54s
-    #   chunk_size = 1000   => 63s
-    size = a.shape[0]
-    chunk_size = 10
-    condition = threading.Condition()
-    chunks = []
+def ndarrays(arrays):
+    """
+    Return a list of NumPy ndarray objects corresponding to the given
+    biggus Array objects.
 
-    def read():
-        for i in range(1, size, chunk_size):
-            chunk = a[i:i + chunk_size].ndarray()
-            with condition:
-                chunks.append(chunk)
-                condition.notify()
-        with condition:
-            chunks.append(None)
-            condition.notify()
+    This can be more efficient (and hence faster) than converting the
+    individual arrays one by one.
 
-    producer = threading.Thread(target=read)
-    producer.start()
-
-    while True:
-        with condition:
-            while not chunks and producer.is_alive():
-                condition.wait(1)
-            chunk = chunks.pop(0)
-        if chunk is None:
-            break
-        chunk_handler(chunk)
+    """
+    # Pick out any _Aggregation nodes which have a simple ArrayAdapter
+    # source.
+    ndarrays = [None] * len(arrays)
+    aggregations_by_src = {}
+    for i, array in enumerate(arrays):
+        if isinstance(array, _Aggregation) and \
+                isinstance(array._array, ArrayAdapter):
+            aggregations = aggregations_by_src.setdefault(array._array, [])
+            aggregations.append((i, array))
+        else:
+            ndarrays[i] = array.ndarray()
+    for aggregations in aggregations_by_src.itervalues():
+        indexes = [agg[0] for agg in aggregations]
+        agg_arrays = [agg[1] for agg in aggregations]
+        agg_ndarrays = _aggregation_ndarrays(agg_arrays)
+        for i, ndarray in zip(indexes, agg_ndarrays):
+            ndarrays[i] = ndarray
+    return ndarrays
 
 
-class Aggregation(Array):
-    def __init__(self, operator, array, axis):
-        self._operator = operator
+def _aggregation_ndarrays(arrays):
+    chunk_handlers = [array.chunk_handler() for array in arrays]
+    for chunk_handler in chunk_handlers:
+        chunk_handler.bootstrap()
+
+    def meta_chunk_handler(chunk):
+        for chunk_handler in chunk_handlers:
+            chunk_handler.add_chunk(chunk)
+
+    src_array = arrays[0]._array
+    _process_chunks(src_array, meta_chunk_handler)
+
+    results = [chunk_handler.result() for chunk_handler in chunk_handlers]
+    return results
+
+
+class _ChunkHandler(object):
+    __metaclass__ = ABCMeta
+
+    def __init__(self, array, axis, kwargs):
+        self.array = array
+        self.axis = axis
+        self.kwargs = kwargs
+
+    @abstractmethod
+    def bootstrap(self):
+        pass
+
+    @abstractmethod
+    def add_chunk(self, chunk):
+        pass
+
+    @abstractmethod
+    def result(self):
+        pass
+
+
+class _Mean(_ChunkHandler):
+    def bootstrap(self):
+        first_slice = self.array[self.axis].ndarray()
+        self.running_total = np.array(first_slice)
+        self.t = np.empty_like(first_slice)
+
+    def add_chunk(self, chunk):
+        np.sum(chunk, axis=self.axis, out=self.t)
+        self.running_total += self.t
+
+    def result(self):
+        self.running_total /= self.array.shape[0]
+        return self.running_total
+
+
+class _Std(_ChunkHandler):
+    # The algorithm used here preserves numerical accuracy whilst only
+    # requiring a single pass, and is taken from:
+    # Welford, BP (August 1962). "Note on a Method for Calculating
+    # Corrected Sums of Squares and Products".
+    # Technometrics 4 (3): 419-420.
+    # http://zach.in.tu-clausthal.de/teaching/info_literatur/Welford.pdf
+
+    def bootstrap(self):
+        first_slice = self.array[self.axis].ndarray()
+        self.a = np.array(first_slice)
+        self.q = np.zeros_like(first_slice)
+        self.t = np.empty_like(first_slice)
+        self.k = 1
+
+    def add_chunk(self, chunk):
+        chunk = np.rollaxis(chunk, self.axis)
+        for slice in chunk:
+            self.k += 1
+
+            # Compute A(k)
+            np.subtract(slice, self.a, out=self.t)
+            self.t *= 1. / self.k
+            self.a += self.t
+
+            # Compute Q(k)
+            self.t *= self.t
+            self.t *= self.k * (self.k - 1)
+            self.q += self.t
+
+    def result(self):
+        assert self.k == self.array.shape[self.axis]
+        self.q /= (self.k - self.kwargs['ddof'])
+        result = np.sqrt(self.q)
+        return result
+
+
+class _Aggregation(Array):
+    def __init__(self, array, axis, chunk_handler_class, kwargs):
         self._array = array
         self._axis = axis
+        self._chunk_handler_class = chunk_handler_class
+        self._kwargs = kwargs
 
     @property
     def dtype(self):
@@ -569,14 +652,18 @@ class Aggregation(Array):
         if not isinstance(keys, tuple):
             keys = (keys,)
         keys = (slice(None),) + keys
-        return Aggregation(self._operator, self._array[keys], self._axis)
+        return _Aggregation(self._array[keys], self._axis,
+                            self._chunk_handler_class, self._kwargs)
 
     def ndarray(self):
-        return self._operator(self._array, axis=self._axis)
+        return ndarrays([self])[0]
 
     def masked_array(self):
         raise RuntimeError()
 
+    def chunk_handler(self):
+        return self._chunk_handler_class(self._array, self._axis, self._kwargs)
+
 
 def mean(a, axis=None):
     """
@@ -585,27 +672,7 @@ def mean(a, axis=None):
     NB. Currently limited to axis=0.
 
     """
-    return Aggregation(_mean, a, axis)
-
-
-def _mean(a, axis=None):
-    """
-    Returns the mean of an Array as a NumPy ndarray.
-
-    NB. Currently limited to axis=0.
-
-    """
-    assert axis == 0
-    size = a.shape[0]
-    total = a[axis].ndarray()
-    t = np.empty_like(total)
-
-    def chunk_handler(chunk):
-        np.sum(chunk, axis=axis, out=t)
-        np.add(total, t, out=total)
-
-    _process_chunks(a, chunk_handler)
-    return np.divide(total, size, out=total)
+    return _Aggregation(a, axis, _Mean, {})
 
 
 def std(a, axis=None, ddof=0):
@@ -615,44 +682,33 @@ def std(a, axis=None, ddof=0):
     NB. Currently limited to axis=0.
 
     """
-    def _operator(array, axis):
-        return _std(array, axis, ddof=ddof)
-    return Aggregation(_operator, a, axis)
+    return _Aggregation(a, axis, _Std, {'ddof': ddof})
 
 
-def _std(a, axis=None, ddof=0):
-    """
-    Return the mean of an Array as a NumPy ndarray.
+def _process_chunks(array, chunk_handler):
+    #   chunk_size = 2      => 54s ~ 115% CPU
+    #   chunk_size = 10     => 42s ~ 105% CPU (quicker than CDO!)
+    #   chunk_size = 100    => 54s
+    #   chunk_size = 1000   => 63s
+    size = array.shape[0]
+    chunk_size = 10
+    chunks = Queue.Queue(maxsize=3)
 
-    NB. Currently limited to axis=0.
+    def worker():
+        while True:
+            chunk = chunks.get()
+            chunk_handler(chunk)
+            chunks.task_done()
 
-    """
-    # NB. This algorithm is not particularly good for numerical accuracy.
-    assert axis == 0
-    assert ddof in (0, 1)
-    total = np.array(a[axis].ndarray())
-    total_of_squares = np.asarray(total * total)
-    t = np.empty_like(total)
+    thread = threading.Thread(target=worker)
+    thread.daemon = True
+    thread.start()
 
-    def chunk_handler(chunk):
-        np.sum(chunk, axis=axis, out=t)
-        np.add(total, t, out=total)
-        # TODO: Might be faster to re-use a chunk-sized scratch array
-        # for computing the squares.
-        np.sum(chunk * chunk, axis=axis, out=t)
-        np.add(total_of_squares, t, out=total_of_squares)
+    for i in range(1, size, chunk_size):
+        chunk = array[i:i + chunk_size].ndarray()
+        chunks.put(chunk)
 
-    _process_chunks(a, chunk_handler)
-    # TODO: Optimise
-    size = np.array(a.shape[0], dtype=total.dtype)
-    if ddof == 0:
-        result = np.sqrt(size * total_of_squares - total * total) / size
-    else:
-        result = np.sqrt((size * total_of_squares - total * total) /
-                         (size * (size - 1)))
-    if result.ndim == 0:
-        result = np.array(result)
-    return result
+    chunks.join()
 
 
 # TODO: Test
